import 'package:dio/dio.dart';
import 'package:localstore/localstore.dart';

import 'bagel_db_functions.dart';

String authEndpoint = "https://auth.bageldb.com/api/public";
class Collcetions{
    static String user ="user";
    static String access ="access";
}
class BagelUsersRequest {
  BagelDB instance;
  late Dio dio;
  final _db = Localstore.instance;
  BagelUsersRequest(this.instance){
    dio = Dio(BaseOptions(headers: {
      'authorization': 'Bearer ${this.instance.token}',
      "Accept-Version": "v1"
    }));
  }

  /// Check if the user is logged in
  Future<bool> isLoggedIn()async {
    return _bagelUserActive();
  }

  Future<bool> _bagelUserActive() async {
    return await getUserID() != null;
  }

/// Create a user with a an email and password
  Future<String> create(email, password) async {
    String url = '$authEndpoint/user';
    Map body = {"email": email, "password": password};
  try  {
  Response res = await dio.post(url, data: body);
      _storeTokens(res.data["access_token"], res.data["refresh_token"], res.data["expires_in"]);
      _storeBagelUser(res.data["user_id"]);
      return res.data["user_id"];
    } catch(err){
      throw(err);
    }
  }

  Future<String> _getOtpRequestNonce() async {
      Map<String,dynamic>? otpRequest = await _db.collection(Collcetions.access)
        .doc("nonce")
        .get();
      if (otpRequest!=null){
          DateTime expires = DateTime.parse(otpRequest["expires"]);
          if (new DateTime.now().isAfter(expires)) throw("OTP request has expired, try again");
          return otpRequest["nonce"];
      } else{
        return "";
      }
  }

  _storeOtpRequestNonce(Map otpRequest) {
    DateTime expires =
        new DateTime.now().add(Duration(minutes: otpRequest["expires_in"]));
    _db
      .collection(Collcetions.access)
      .doc("nonce")
      .set({"nonce": otpRequest["nonce"], "expires": expires.toString()});
  }

  /// Request the user receive a One Time Password either via SMS or Email
  Future<String> requestOtp(emailOrPhone) async {
    String url = '$authEndpoint/user/otp';
    Map body = {"emailOrPhone": emailOrPhone};

    Response res = await dio.post(url, data: body);
    _storeOtpRequestNonce(res.data);
    return res.data["nonce"];
  }

  /// Validate the user's received One Time Password
  Future<String> validateOtp(String otp) async {
    String nonce = await _getOtpRequestNonce();
    String url = '$authEndpoint/user/otp/verify/$nonce';
    Map body = {"otp":otp};
    Response res = await dio.post(url, data: body);

    _storeTokens(res.data["access_token"], res.data["refresh_token"], res.data["expires_in"]);
    _storeBagelUser(res.data["user_id"]);
    return res.data["user_id"];
  }

  /// Authenticate the user with an email and password
  Future<String> validate(email, password) async {
    String url = '$authEndpoint/user/verify';
    Map body = {"email": email, "password": password};

    Response res = await dio.post(url, data: body);
    if (res.statusCode == 200 || res.statusCode == 201){
        _storeTokens(res.data["access_token"], res.data["refresh_token"], res.data["expires_in"]);
        _storeBagelUser(res.data["user_id"]);
        return res.data["user_id"];
    } else{
      throw("wrong authorization codes");
    }

  }

  /// Get the currently logged in user's info
  getUser() async {
    if (!await _bagelUserActive()) {
      throw ("a Bagel User must be logged in to get Bagel User info");
    }
    String? accessToken = await getAccessToken();
     Options options =  Options(headers: {
      'authorization': 'Bearer $accessToken',
      "Accept-Version": "v1"
    });
    String url = '$authEndpoint/user';
    Response res = await dio.get(url,options: options);
    return res.data;
  }

  _storeBagelUser(userID) {
    _db.collection(Collcetions.user).doc("userID").set({"userID": userID});
  }

  _storeTokens(accessToken, refreshToken, expires) {
    _db.collection(Collcetions.access)
      .doc("access")
      .set({
        "expiryTime": DateTime.now().add(Duration(seconds: expires)).toString(),
        "accessToken": accessToken,
        "refreshToken": refreshToken
      });
    }

  /// The currently logged in user's ID
  Future<String?> getUserID() async {
    
    Map<String, dynamic>? userID =
        await _db.collection(Collcetions.user).doc("userID").get();
    return userID?["userID"];
  }

  Future<String?> _getRefreshToken() async {
    Map<String, dynamic>? refreshToken =
        await _db.collection(Collcetions.access).doc("access").get();
    return refreshToken?["refreshToken"];
  }

  Future<String?>getAccessToken() async {
    Map<String, dynamic>? accessToken =
        await _db.collection(Collcetions.access).doc("access").get();
    return accessToken?["accessToken"];
  }

  /// Log the user out
  void logout() async {
    await _db.collection(Collcetions.access).doc("access").delete();
    await _db.collection(Collcetions.user).doc("userID").delete();
  }

  /// is used to refresh the users auth token without having to login with username and password
  refresh() async {
    if (await _getRefreshToken() != null) {
      throw ("No Bagel User is logged in");
    }
    String url = '$authEndpoint/user/token';
    String? refreshToken = await _getRefreshToken();
    String body =
        'grant_type=refresh_token&refresh_token=$refreshToken&client_id=project-client';
    Response res = await dio.post(url, data: body);
    if (res.statusCode == 200) {
      _storeTokens(res.data["access_token"], res.data["refresh_token"], res.data["expires_in"]);
      return res;
    } else {
      throw (res);
    }
  }
}
